/*  This file is part of the simple Vector Field Plotter (vfplot) project

    Written by Nicolas Brodu <nicolas.brodu@numerimoire.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef IMAGE_PRODUCER
#define IMAGE_PRODUCER

#include <fstream>
#include <vector>
#include <array>
#include <limits>
#include <algorithm>
#include <string>

#include <boost/format.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/variate_generator.hpp>
#include <boost/random/uniform_01.hpp>
#include <boost/operators.hpp>

#include <stdint.h>
#include <cairo.h>
#include <cairo-svg.h>

#include <math.h>

// Windows Hacks needed on that "OS"
#ifdef WHACK
#include <sstream>
#include <png++/png.hpp>
#include "base64.hpp"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

inline double exp10(double x) {
    return exp(log(10.)*x);
}

struct pngpp_ostream {
    std::vector<char> data;
    void write(char const* c, size_t s) {
        for (int i=0; i<s; ++i) data.push_back(c[i]);
    }
    void flush() {}
    bool good() {return true;}
};
 
#endif

struct ImageProducer {
    
    // hacked down version of the vec2/3/quaternion classes in my other projects
    struct vec2 : boost::addable<vec2, boost::subtractable<vec2, boost::multipliable2<vec2, double, boost::dividable2<vec2, double> > > > {
        double x,y;
        inline vec2(double _x, double _y) : x(_x),y(_y) {}
        inline vec2() : x(0),y(0) {}
        // Thanks boost::operators for easing the task
        inline vec2& operator+=(const vec2& v) {x+=v.x; y+=v.y; return *this;}
        inline vec2& operator-=(const vec2& v) {x-=v.x; y-=v.y; return *this;}
        inline vec2& operator*=(const double& f) {x*=f; y*=f; return *this;}
        inline vec2& operator/=(const double& f) {return *this *= 1.0/f;}
        inline vec2& operator=(const double& f) {x=f; y=f; return *this;}
        inline double norm2() const {return x*x+y*y;}
        inline double normalize() {
            double n = norm2();
            if (n<=0) return 0;
            n = sqrt(n);
            (*this) *= 1.0/n;
            return n;
        }
    };
    
    int width, height;
    std::vector<double> data;
    std::vector<double> lightness; // allocated on demand
    std::vector<vec2> vf;          // vector field, allocated on demand
    std::vector<double> vfnorm;
    double data_min, data_max;
    double x_per_pixel, y_per_pixel;
    double font_size;
    double x0, y0;
    double arrows_range_min, arrows_range_max;
    
    ImageProducer(int _width, int _height, double _x_per_pixel, double _y_per_pixel, double _x0 = 0., double _y0 = 0.) : 
        width(_width), 
        height(_height), 
        data(_width * _height, 0), 
        data_min(std::numeric_limits<double>::max()), 
        data_max(-std::numeric_limits<double>::max()),
        x_per_pixel(_x_per_pixel), y_per_pixel(_y_per_pixel),
        x0(_x0), y0(_y0) {
            font_size = 10.0;
            arrows_range_min = arrows_range_max = -1;
    }
    
    void plot(int x, int y, double value) {
        // column-major to match loops in main program that produce text files
        // compatible with octave
        data[x*height+y] = value;
        data_min = std::min(value, data_min);
        data_max = std::max(value, data_max);
    }
    
    void plot(int x, int y, double value, double light) {
        lightness.resize(data.size(),0); // allocate/initialises on first call only
        lightness[x*height+y] = std::min(1.,std::max(0.,light));
        plot(x,y,value);
    }

    // vector field visualization with fast line integral convolution
    void vfplot(int x, int y, double vx, double vy) {
        vf.resize(data.size());
        vfnorm.resize(data.size());
        // separate direction from norm
        vf[x*height+y] = vec2(vx,vy);
        vfnorm[x*height+y] = vf[x*height+y].normalize();
        plot(x,y, vfnorm[x*height+y]);
    }

    // vector field as just arrows on top of a scalar field
    void vfplot(int x, int y, double vx, double vy, double value) {
        vf.resize(data.size());
        vfnorm.resize(data.size());
        // separate direction from norm
        vf[x*height+y] = vec2(vx,vy);
        vfnorm[x*height+y] = vf[x*height+y].normalize();
        plot(x,y, value);
    }

    void set_arrows_range(double _arrows_range_min, double _arrows_range_max) {
        arrows_range_min = _arrows_range_min;
        arrows_range_max = _arrows_range_max;
    }

    // may bypass min/max, ex: for having the same scale across all simu runs
    void save(const std::string& filename, double dmin, double dmax) {
        data_min = dmin;
        data_max = dmax;
        save(filename);
    }
    
    void save(const std::string& filename) {
        using namespace std;
        using namespace boost;
        
        if (!vf.empty()) set_lightness_with_lic();
        
        double delta_data = data_max - data_min;
        if (delta_data==0) delta_data = 1;

        vector<uint8_t> argbdata(width * height * 4);
        cairo_surface_t* image_surface = cairo_image_surface_create_for_data(&argbdata[0], CAIRO_FORMAT_ARGB32, width, height, width*4); // width*4 = stride in bytes
        cairo_t* cr = cairo_create (image_surface);
        cairo_surface_flush(image_surface);
        // Pixels can be modified directly between flush and mark_dirty
        for (int py=0; py<height; ++py) for (int px=0; px<width; ++px) {
            // clamp the scale value as user may set arbirtary min/max
            double s = min(1., max(0., (data[px*height+py] - data_min) / delta_data));
            double r,g,b;
            if (lightness.empty()) palette(s,r,g,b);
            else palette(s,lightness[px*height+py],r,g,b);
            // correctly place the bottom on the bottom... flip now
            int offset = ((height-1-py) * width + px) * 4;
            argbdata[offset + 0] = (int)floor(b*(256-1e-6)); // blue
            argbdata[offset + 1] = (int)floor(g*(256-1e-6)); // green
            argbdata[offset + 2] = (int)floor(r*(256-1e-6)); // red
            argbdata[offset + 3] = 255;   // alpha
        }
        cairo_surface_mark_dirty(image_surface);
        //cairo_surface_write_to_png(image_surface,(output_directory+"/packing_fraction_image.png").c_str());
        
        double xshift,yshift;
        double xtick = auto_ticks(cr,x0,x0+x_per_pixel*width,width,false,xshift);
        int nxticks = (int)ceil(x_per_pixel*width/xtick);
        if (fabs(nxticks*xtick / (x_per_pixel*width)-1.) < 1e-6) ++nxticks;
        double ytick = auto_ticks(cr,y0,y0+y_per_pixel*height,height,true,yshift);
        int nyticks = (int)ceil(y_per_pixel*height/ytick);
        if (fabs(nyticks*ytick / (y_per_pixel*height)-1.) < 1e-6) ++nyticks;
        
        cairo_text_extents_t extents;
        cairo_select_font_face (cr, "Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
        cairo_set_font_size (cr, font_size);
        string s = str(format("0%g0") % ((nxticks-1) * xtick));
        cairo_text_extents (cr, s.c_str(), &extents);
        double maxtextright = (nxticks-1) * xtick / x_per_pixel + extents.width/2 - extents.x_bearing;
        double mintextleft = 0;
        for (int i=0; i<nyticks; ++i) {
            s = str(format("0%g") % (y0 + i * ytick));
            cairo_text_extents (cr, s.c_str(), &extents);
            mintextleft = max(mintextleft, extents.width + extents.x_bearing + font_size);
        }
        
        cairo_surface_t* svgsurface = cairo_svg_surface_create((filename+".svg").c_str(), max((double)width,maxtextright)+mintextleft, height+font_size*3);
        cairo_t* crsvg = cairo_create(svgsurface);
        cairo_select_font_face (crsvg, "Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
        cairo_set_font_size(crsvg, font_size);

        cairo_set_source_rgb(crsvg, 0, 0, 0);
        cairo_set_line_width (crsvg, 1);
        for (int i=0; i<nxticks; ++i) {
            s = str(format("%g") % round5digit(x0 + i * xtick));
            cairo_text_extents (crsvg, s.c_str(), &extents);
            double midx = mintextleft + i * xtick / x_per_pixel;
            cairo_move_to(crsvg, midx - (extents.width/2 + extents.x_bearing), height + font_size*2.5);
            cairo_show_text(crsvg, s.c_str());
            cairo_move_to(crsvg, midx, height + font_size*0.5);
            cairo_line_to(crsvg, midx, height + font_size);
            cairo_stroke(crsvg);
        }
        for (int i=0; i<nyticks; ++i) {
            s = str(format("%g") % round5digit(y0 + i * ytick));
            cairo_text_extents(crsvg, s.c_str(), &extents);
            double midy = height + font_size*0.5 - (i * ytick / y_per_pixel);
            cairo_move_to(crsvg, mintextleft - font_size - extents.width, midy - (extents.height/2 + extents.y_bearing));
            cairo_show_text(crsvg, s.c_str());
            cairo_move_to(crsvg, mintextleft - font_size*0.5, midy);
            cairo_line_to(crsvg, mintextleft, midy);
            cairo_stroke(crsvg);
        }
        
        cairo_translate(crsvg, mintextleft, font_size*0.5);
        
        // PNG support is badly broken, skip it and hack to put back image below
#ifndef WHACK
        cairo_set_source_surface (crsvg, image_surface, 0, 0);
        cairo_paint (crsvg);
#endif
        cairo_set_source_rgb(crsvg, 0, 0, 0);
        cairo_set_line_width (crsvg, 1);
        cairo_move_to(crsvg, 0, height);
        cairo_line_to(crsvg, 0, 0);
        cairo_move_to(crsvg, 0, height);
        cairo_line_to(crsvg, width, height);
        cairo_move_to(crsvg, width, height);
        cairo_line_to(crsvg, width, 0);
        cairo_stroke(crsvg);

        // trace vector field arrows if requested
        if (!vf.empty()) {
            // median arrow is half the distance between arrow grid points, so 0.25 tick
            double avg_arrow_len = 0.25 * min((xtick / x_per_pixel), (ytick / y_per_pixel));
            // max is full grid
            double max_arrow_len = 2*avg_arrow_len;
            double davg, nmin;

            if (arrows_range_min>=0) {
                davg = (arrows_range_min + arrows_range_max) * 0.5;
                nmin = arrows_range_min;
            } else {
                vector<double> normcopy = vfnorm;
                sort(normcopy.begin(), normcopy.end());
                // median value EXCLUDING null values...
                int didx = 0; while (didx<normcopy.size() && normcopy[didx]==0) ++didx;
                davg = normcopy[(didx+data.size())/2];
                nmin = normcopy.front();
            }

            // head shafts at 30° angle from the main direction
            const double cost = sqrt(3.0) * 0.5; 
            const double sint = 0.5; 
            for (int i=1; i<2*nxticks; ++i) {
                int x = (int)(i * 0.5 * xtick / x_per_pixel); 
                if (x>=width) continue;
                for (int j=1; j<2*nyticks; ++j) {
                    int y = (int)(j * 0.5 * ytick / y_per_pixel);
                    if (y>=height) continue;
                    
                    // curved arrows are cool !
                    double streamlen = 0;
                    const double step = 0.5;
                    double maxlen = min(max_arrow_len, avg_arrow_len * (vfnorm[x*height+y] - nmin) / (davg - nmin));
                    if (maxlen<=0) continue; // null data < dmin
                    double davg_path = vfnorm[x*height+y];
                    int npath = 1;
                    cairo_move_to(crsvg, x, height-1-y);
                    
                    vec2 position(x,y); vec2 lastpos = position;
                    position += vf[x*height+y] * step;
                    vec2 vend;
                    while (streamlen<=maxlen) {
                        int x = (int)floor(position.x);
                        int y = (int)floor(position.y);
                        if (x<0 || x>=width || y<0 || y>=height) break;
                        davg_path += vfnorm[x*height+y]; ++npath;
                        cairo_line_to(crsvg, position.x, height-1-position.y);
                        vend = lastpos - position;
                        lastpos = position;
                        if (streamlen==maxlen) break;
                        position += vf[x*height+y] * step;
                        streamlen += step;
                        if (streamlen>maxlen) {
                            position = lastpos + vf[x*height+y] * (maxlen + step - streamlen);
                            streamlen = maxlen;
                        }
                    }
                    vend.normalize();
                    // arrow head around lastpos
                    // shaft length proportional to total length, say 1/4
                    maxlen /= 4.;
                    if (maxlen < 2) maxlen = 2; // pixels
                    if (maxlen>max_arrow_len) maxlen = max_arrow_len;
                    // rotation matrix multiplication
                    vec2 rotated;
                    rotated.x = cost * vend.x - sint * vend.y;
                    rotated.y = sint * vend.x + cost * vend.y;
                    vec2 shaftend = lastpos + rotated * maxlen;
                    cairo_move_to(crsvg, shaftend.x, height-1-shaftend.y);
                    cairo_line_to(crsvg, lastpos.x, height-1-lastpos.y);
                    // rotate in the other direction
                    rotated.x =  cost * vend.x + sint * vend.y;
                    rotated.y = -sint * vend.x + cost * vend.y;
                    shaftend = lastpos + rotated * maxlen;
                    cairo_line_to(crsvg, shaftend.x, height-1-shaftend.y);
                    // draw all lines, at last, with the average color along the line
                    davg_path /= npath;
                    double r,g,b;
                    palette(davg_path, 0.1, r,g,b);
                    cairo_set_source_rgb(crsvg, r, g, b);
                    cairo_stroke(crsvg);
                }
            }
        }

        cairo_surface_destroy (image_surface);
        cairo_destroy (cr);
        
        cairo_surface_finish (svgsurface);
        cairo_surface_destroy (svgsurface);
        cairo_destroy (crsvg);

        // horrendous windows hack. Bug in cairo, image not generated
#ifdef WHACK
        png::image<png::rgb_pixel> image(width, height);
        for (size_t y = 0; y < image.get_height(); ++y) {
            for (size_t x = 0; x < image.get_width(); ++x) {
                int offset = (y * width + x) * 4;
                image[y][x] = png::rgb_pixel(argbdata[offset + 2], argbdata[offset + 1], argbdata[offset + 0]);
            }
        }
        pngpp_ostream pngsstream;
        image.write_stream(pngsstream);
        vector<char>& pngdata = pngsstream.data;
        
        base64 codec;
        vector<char> base64pngdata(codec.get_max_encoded_size(pngdata.size()));
        codec.reset_encoder();
        int nbytes = codec.encode(&pngdata[0], pngdata.size(), &base64pngdata[0]);
        nbytes += codec.encode_end(&base64pngdata[nbytes]);
        
        ifstream svgfile((filename+".svg").c_str());
        string sline;
        vector<std::string> lines;
        bool first_transform = true;
        while (svgfile && !svgfile.eof()) {
            getline(svgfile, sline);
            if (first_transform) {
                int pos = sline.find("transform");
                if (pos!=-1) {
                    lines.push_back(str(format("<image width=\"%d\" height=\"%d\" xlink:href=\"data:image/png;base64,%s\" %s")
                    % width % height % &base64pngdata[0] % sline.substr(pos)));
                    lines.push_back(sline);
                    first_transform = false;
                }
            }
            lines.push_back(sline);
        }
        svgfile.close();
        ofstream osvgfile((filename+".svg").c_str());
        for (auto l : lines) osvgfile << l << endl;
        osvgfile.close();
#endif
        
        
        // Make the legend

        const int lwidth = 40, lheight = 200;
        argbdata.resize(lwidth * lheight * 4);
        image_surface = cairo_image_surface_create_for_data(&argbdata[0], CAIRO_FORMAT_ARGB32, lwidth, lheight, lwidth*4);
        cr = cairo_create (image_surface);
        cairo_surface_flush(image_surface);
        for (int py=0; py<lheight; ++py) {
            double s = py/(lheight-1.);
            double r,g,b;
            if (lightness.empty()) palette(s,r,g,b);
            else palette(s,0.5,r,g,b);
            int baseoffset = (lheight-1-py) * lwidth;
            for (int px=0; px<lwidth; ++px) {
                int offset = (baseoffset + px) * 4;
                argbdata[offset + 0] = (int)floor(b*(256-1e-6)); // blue
                argbdata[offset + 1] = (int)floor(g*(256-1e-6)); // green
                argbdata[offset + 2] = (int)floor(r*(256-1e-6)); // red
                argbdata[offset + 3] = 255;   // alpha
            }
        }
        cairo_surface_mark_dirty(image_surface);
        ytick = auto_ticks(cr,data_min, data_max,lheight,true,yshift);
        nyticks = (int)ceil(delta_data/ytick);
        if (fabs(nyticks*ytick / delta_data -1.) < 1e-6) ++nyticks;
        
        cairo_select_font_face (cr, "Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
        cairo_set_font_size (cr, font_size);
        mintextleft = 0;
        for (int i=0; i<nyticks; ++i) {
            s = str(format("0%g") % (i * ytick + yshift));
            cairo_text_extents (cr, s.c_str(), &extents);
            mintextleft = max(mintextleft, extents.width + extents.x_bearing + font_size);
        }

        svgsurface = cairo_svg_surface_create((filename+"_legend.svg").c_str(), lwidth+mintextleft+font_size*0.5, lheight+font_size);
        crsvg = cairo_create(svgsurface);
        cairo_select_font_face (crsvg, "Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
        cairo_set_font_size(crsvg, font_size);

        cairo_set_source_rgb(crsvg, 0, 0, 0);
        cairo_set_line_width (crsvg, 1);
        for (int i=0; i<nyticks; ++i) {
            s = str(format("%g") % round5digit(i * ytick + yshift));
            cairo_text_extents(crsvg, s.c_str(), &extents);
            double midy = lheight + font_size*0.5 - ((i*ytick+yshift-data_min) * lheight / delta_data);
            cairo_move_to(crsvg, mintextleft - font_size - extents.width, midy - (extents.height/2 + extents.y_bearing));
            cairo_show_text(crsvg, s.c_str());
            cairo_move_to(crsvg, mintextleft - font_size*0.5, midy);
            cairo_line_to(crsvg, mintextleft, midy);
            cairo_stroke(crsvg);
        }
        cairo_translate(crsvg, mintextleft, font_size*0.5);
        
        // PNG support is badly broken, skip it and hack to put back image below
#ifndef WHACK
        cairo_set_source_surface (crsvg, image_surface, 0, 0);
        cairo_paint (crsvg);
#endif

        cairo_set_source_rgb(crsvg, 0, 0, 0);
        cairo_set_line_width (crsvg, 1);
        cairo_move_to(crsvg, 0, 0);
        cairo_line_to(crsvg, 0, lheight);
        cairo_line_to(crsvg, lwidth, lheight);
        cairo_line_to(crsvg, lwidth, 0);
        cairo_line_to(crsvg, 0, 0);
        cairo_stroke(crsvg);

        cairo_destroy (cr);
        cairo_surface_destroy (image_surface);
        
        cairo_destroy (crsvg);
        
        cairo_surface_destroy (svgsurface);
        
        // horrendous windows hack. Bug in cairo, image not generated
#ifdef WHACK
        {
        png::image<png::rgb_pixel> image(lwidth, lheight);
        for (size_t y = 0; y < image.get_height(); ++y) {
            for (size_t x = 0; x < image.get_width(); ++x) {
                int offset = (y * lwidth + x) * 4;
                image[y][x] = png::rgb_pixel(argbdata[offset + 2], argbdata[offset + 1], argbdata[offset + 0]);
            }
        }
        pngpp_ostream pngsstream;
        image.write_stream(pngsstream);
        vector<char>& pngdata = pngsstream.data;
        
        base64 codec;
        vector<char> base64pngdata(codec.get_max_encoded_size(pngdata.size()));
        codec.reset_encoder();
        int nbytes = codec.encode(&pngdata[0], pngdata.size(), &base64pngdata[0]);
        nbytes += codec.encode_end(&base64pngdata[nbytes]);
        
        ifstream svgfile((filename+"_legend.svg").c_str());
        string sline;
        vector<std::string> lines;
        bool first_transform = true;
        while (svgfile && !svgfile.eof()) {
            getline(svgfile, sline);
            if (first_transform) {
                int pos = sline.find("transform");
                if (pos!=-1) {
                    lines.push_back(str(format("<image width=\"%d\" height=\"%d\" xlink:href=\"data:image/png;base64,%s\" %s")
                    % lwidth % lheight % &base64pngdata[0] % sline.substr(pos)));
                    lines.push_back(sline);
                    first_transform = false;
                }
            }
            lines.push_back(sline);
        }
        svgfile.close();
        ofstream osvgfile((filename+"_legend.svg").c_str());
        for (auto l : lines) osvgfile << l << endl;
        osvgfile.close();
        }
#endif
        
    }
    
    // avoid displaying roundup errors like 1e-18 instead of 0
    inline double round5digit(double x) {
        // use next power of 2 to avoid remaining rounding errors ?
        return round(x * 1e5) * 1e-5;
    }
    
    // return a correct tick for scaling data range [xmin..xmax] into w pixels
    // returns the tick, sets the xshift for the first tick
    inline double auto_ticks(cairo_t* cr, double xmin, double xmax, int w, bool is_vertical, double& xshift) {
        // pixel range at least sufficient for holding text values
        // start with at least 4 characters
        cairo_select_font_face (cr, "Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
        cairo_set_font_size (cr, font_size);
        std::string ws = boost::str(boost::format("0%g0") % w);
        cairo_text_extents_t extents;
        cairo_text_extents (cr, ws.c_str(), &extents);
        int maxnticks = (int)floor(w / extents.width);
        if (is_vertical) maxnticks = (int)floor(w / (3*font_size));
        double x = xmax - xmin;
        double p = exp10(floor(log10(x / maxnticks)));
        int t = (int)floor((x / maxnticks) / p);
        if (t>5) t=5;
        double tick = t*p;
        xshift = ceil(xmin/tick)*tick;
        return tick;
    }
    
    static inline void hue_to_rgb(double hue, double& r, double& g, double& b) {
        hue = 6 * (hue - floor(hue)); // 0 <= hue < 1
        switch ((int)hue) {
            case 0: r=1; b=0; g = hue; return;
            case 1: g=1; b=0; r = 2-hue; return;
            case 2: g=1; r=0; b = hue-2; return;
            case 3: b=1; r=0; g = 4-hue; return;
            case 4: b=1; g=0; r = hue-4; return;
            case 5: r=1; g=0; b = 6-hue; return;
        }
    }
    
    static inline void palette(double x, double l, double& r, double& g, double& b) {
        // arc-circle in the UV plane. from 360+15° down to 90-15°
        static const double amin = M_PI * (2.);//+1./6.);
        static const double amax = M_PI * (0.5-1./6.);
        double a = amin + x * (amax - amin);
        double u = cos(a) * 0.5 * sin(M_PI * l);
        double v = sin(a) * 0.5 * sin(M_PI * l);
        // much better / smoother gray tones using the square when printing
        double y = l*l;
        // this is _not_ the usual yuv->rgb. 
        // It was obtained by inverting y = 0.28000 r + 0.62000 g + 0.10000 b
        // and the same lines as usual for u,v.
        // Thus green is given even more weight, matching observations
        // on a grayscale printer.
        r = std::min(1., std::max(0., y + 0.041430 * u + 1.180634 * v));
        g = std::min(1., std::max(0., y - 0.353141 * u - 0.539694 * v));
        b = std::min(1., std::max(0., y + 2.073469 * u + 0.040325 * v));
    }

    // with caching
    static inline void palette(double value, double& r, double& g, double& b) {
        static const int N = 1000;
        // allocate on first call idiom, n+1 entries
        static std::array<double,3>* cache = create_palette_cache(N);
        int idxlow = (int)floor(value*N);
        double rem = value*N - idxlow;
        int idxhigh = (int)ceil(value*N);
        if (idxhigh<=0) {r=cache[0][0]; g=cache[0][1]; b=cache[0][2];}
        if (idxlow>=N) {r=cache[N][0]; g=cache[N][1]; b=cache[N][2];}
        r = cache[idxlow][0] + (cache[idxhigh][0] - cache[idxlow][0]) * rem;
        g = cache[idxlow][1] + (cache[idxhigh][1] - cache[idxlow][1]) * rem;
        b = cache[idxlow][2] + (cache[idxhigh][2] - cache[idxlow][2]) * rem;
    }

    static inline std::array<double,3>* create_palette_cache(int N) {
        std::array<double,3>* ret = new std::array<double,3>[N+1];
        for (int i=0; i<=N; ++i) {
            // from 0 to 1 included
            double x = (double)i/(double)N;
            // store the rgb pattern in the cache
            palette(x,1-x,ret[i][0],ret[i][1],ret[i][2]);
        }
        return ret;
    }
    
    // this is not a periodic window function, not a good low-pass filter either
    // so do not try to animate the lic with this.
    // It is however a good enough window function for weighting, and fast.
    static inline double half_window(double x, double scale) {
        // use parabola, max at 0, 0 at scale
        return (scale-x)*(scale+x);//scale*scale-x*x;
    }

    // fast streamline approximation using fixed half-pixel steps
    // do not loose too much compared to original lic around singularities, 
    // and _much_ faster
    inline double followflow(double& texsum, vec2 position, const std::vector<double>& texture, bool negate, double maxlen) {
        
        double sumweights = 0;
        double streamlen = 0;
        const double abstep = 0.5;
        const double step = negate ? -abstep : abstep;

        int x = (int)floor(position.x);
        int y = (int)floor(position.y);
        position += vf[x*height+y] * step;
        
        while (streamlen<maxlen) {
            int x = (int)floor(position.x);
            int y = (int)floor(position.y);
            if (x<0 || x>=width || y<0 || y>=height) break;
            double weight = half_window(streamlen,maxlen);
            sumweights += weight;
            texsum += texture[x*height+y] * weight;
            position += vf[x*height+y] * step;
            streamlen += abstep;
        }
        return sumweights;
    }
    
    boost::mt19937* random_source() {
        // construct on first use idiom
        static boost::mt19937* random_source = new boost::mt19937(42);
        return random_source;
    }

    // uses the random seed provided in the options for reproducibility
    // returns a number between 0 included and 1 excluded
    double rand01() {
        static boost::variate_generator<boost::mt19937&, boost::uniform_01<double> >* randgen = new boost::variate_generator<boost::mt19937&, boost::uniform_01<double> >(*random_source(), boost::uniform_01<double>() );
        return (*randgen)();
    }

    // sets the lightness map using line integral convolution from
    // the vector field
    void set_lightness_with_lic() {
        lightness.resize(data.size());
        // first generate a random texture field
        boost::mt19937 random_source(42);
        boost::variate_generator<boost::mt19937&, boost::uniform_01<double> > rand01(random_source, boost::uniform_01<double>() );
        std::vector<double> texture(data.size());
        for (int i=0; i<data.size(); ++i) {
            // pre-process noise in a sigmoid for better contrast
            double r = rand01() * 2 - 1;
            r = r / (1 + fabs(r)) * 2;
            texture[i] = (r + 1)*0.5;
        }
        const double maxlen = 10.; // in pixels
        for (int x=0; x<width; ++x) for (int y=0; y<height; ++y) {
            double sumweights = half_window(0,maxlen);
            double light = texture[x*height+y] * sumweights;
            if (data[x*height+y]==0) {
                light = 1; // line converges here
            } else {
                vec2 position(x+0.5, y+0.5);
                // follow the vector field in the positive then negative directions
                sumweights += followflow(light, position, texture, false, maxlen);
                sumweights += followflow(light, position, texture, true, maxlen);
                light /= sumweights;
            }
            // reserve 0.2 above black so the arrows are more visible
            lightness[x*height+y] = std::min(1., std::max(0., light)) * 0.8 + 0.2;
        }
    }
    
};


#endif
