/*  This file is part of the simple Vector Field Plotter (vfplot) project

    Written by Nicolas Brodu <nicolas.brodu@numerimoire.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <set>
#include <limits>
#include <algorithm>

#include <boost/program_options.hpp>
#include <boost/tokenizer.hpp>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "image_producer.hpp"
#include "fast_atof_next_token.h"

// C I/O routines + our own real number converter are much faster !
#include <stdlib.h>

// Windows Hacks needed to compile on that "OS"
#ifdef WHACK
// implement the spec with minimal compliance for our needs...
ssize_t getline(char **lineptr, size_t *n, FILE *stream) {
    static const int chunck = 80;
    if (*lineptr==0) {
        *lineptr = (char*)malloc(chunck);
        *n = chunck;
    }
    int offset = 0;
    while (true) {
        int c = fgetc(stream);
        if (c==EOF) return -1;
        if (offset>=*n) {
            *n += chunck;
            *lineptr = (char*)realloc(*lineptr,*n);
            if (!*lineptr) return -1; // out of mem, shall set errno...
        }
        (*lineptr)[offset++]=(char)c;
        if (c=='\n') break;
    }
    // terminal 0
    if (offset>=*n) {
        *n += chunck;
        *lineptr = (char*)realloc(*lineptr,*n);
        if (!*lineptr) return -1; // out of mem, shall set errno...
    }
    (*lineptr)[offset]=0;
    return offset;
}
#endif

using namespace std;
namespace po = boost::program_options;

int main(int argc, char** argv) {

    string configfilename, datafilename, outputfilename, fields, image_size, xrange, yrange, scalar_range, arrows_range;
    bool outer_boundaries = false;
    int zoomfactor = 1;
    
    po::options_description optionsdesc("Vector (and scalar) field plotter.\nUsage: vfplot [--options args ...] data_file output_base_name\n\nOptions can be specified in any order");
    optionsdesc.add_options()
        ("data,d", po::value(&datafilename)->default_value(""), "The data file, which can also be specified as the first free parameter on the command line (no need for the --data option). This is a simple ascii multi-column file, see --using. Use the special name - for stdin.")
        ("output,o", po::value(&outputfilename)->default_value(""), "The output files base name, which can also be specified as the second free parameter on the command line (no need for the --output option). Two SVG files are generated, one is base_name.svg for the generated color map / vector field, and the other is base_name_legend.svg for the associated legend. The size of the generated image is set to the number of distinct X and Y values. If this does not suit your needs you may easily rescale the SVG file in a SVG editor (tip: try Inkscape), or use the zoom factor (see below). If this is not enough you'll have to resample your data file before passing it to this program.")
        ("using,u", po::value(&fields)->default_value(""), "Specify which columns to use for plotting, default is 1:2:3 for a 3-column file, 1:2:3:4 for a fourth-column file, and 1:2:3:4:5 otherwise. The first value indicate which column holds the X coordinates, the second value the Y coordinates. The program assumes the X and Y data are regularly spaced, possibly with missing values. If only three values are specified a scalar color map is built. If 4 values are specified a vector field is built using the third and fourth values as the vector coordinates at each X,Y position, and the color is set according to the vector norm. If a fifth value is given then it is used for building a scalar color map, while fields 3 and 4 are still used as vector coordinates for plotting the arrows on top of that map.")
        ("xrange,x", po::value(&xrange)->default_value(""), "The range of the X values, specified as min:max. The default is to use the min and max in the file. Data values outside this range are ignored.")
        ("yrange,y", po::value(&yrange)->default_value(""), "The range of the Y values, specified as min:max. The default is to use the min and max in the file. Data values outside this range are ignored.")
        ("scalar_range,r", po::value(&scalar_range)->default_value(""), "The range of the scalar data, specified as min:max. The default is to use the min and max in the file, but you can specify a different range, for example to have comparable colors across different figures. Values outside the range are clamped to the min/max.")
        ("arrows_range,a", po::value(&arrows_range)->default_value(""), "The range of the arrow lengths for the vector data, specified as min:max. The default is to set the arrow of a median normed vector to half the grid size. Here min and max specify the norm below which no arrow is displayed, and the norm above which the arrow is clamped to the grid size, with the average set to half the grid size. This option is mainly useful to have comparable arrow lengths across different figures")
        ("boundary_values,b", "There is one pixel per value, regularly spaced. Default is to set each value at a pixel center. If this option is set, the data min and max correspond to the left/low boundaries of the first and last pixels.")
        ("zoom,z", po::value(&zoomfactor)->default_value(1,""), "If set, each value in the data file is duplicated that many times along the X and Y directions. This effectively enlarges pixels. Only integer values are allowed, no interpolation is performed, see --output above.")
        ("config,c", po::value(&configfilename)->default_value(""), "If specified, add all parameters from this configuration file in addition to the command-line arguments. The file syntax is one option=value per line, # are comments and blank lines are ignored. In case of duplicated values, the command-line arguments take precedence.")
        ("silent,s", "Do not display warning messages.")
        ("help,h","Display this help")
    ;
    
    po::positional_options_description positionaloptions;
    positionaloptions.add("data", 1).add("output", 1);
    
    po::variables_map args;
    try {
        po::parsed_options parsedopts = po::command_line_parser(argc, argv).              options(optionsdesc).positional(positionaloptions).allow_unregistered().run();
        vector<string> unrecopts = po::collect_unrecognized(parsedopts.options, po::exclude_positional);
        if (!unrecopts.empty()) {
            cout << "Warning: Unrecognized options:" << endl;
            for (auto opt : unrecopts) cout << "  " << opt << endl;
        }
        po::store(parsedopts, args);
        po::notify(args);
    } catch (exception& e) {
        cout << optionsdesc << endl;
        cout << e.what() << endl;
        return 1;
    }

    if (!configfilename.empty()) {
        ifstream configfile(configfilename.c_str());
        if (!configfile) {
            cout << "Warning: Cannot find config file: " << configfilename << endl;
        } else {
            try {
                po::parsed_options parsedopts = po::parse_config_file(configfile, optionsdesc, true);
                vector<string> unrecopts = po::collect_unrecognized(parsedopts.options, po::exclude_positional);
                if (!unrecopts.empty()) {
                    cout << "Warning: Unrecognized options:" << endl;
                    for (auto opt : unrecopts) cout << "  " << opt << endl;
                }
                po::store(parsedopts, args);
                po::notify(args);
            } catch (exception& e) {
                cout << optionsdesc << endl;
                cout << e.what() << endl;
                return 1;
            }
            configfile.close();
        }
    }
    
    if (args.count("help")) {
        cout << optionsdesc << endl;
        return 0;
    }

    bool verbose = true;
    if (args.count("silent")) verbose = false;
    if (args.count("boundary_values")) outer_boundaries = true;

    if (outputfilename.empty()) {
        cerr << optionsdesc << endl;
        cerr << "Error: No output file specified!" << endl;
        return 1;
    }
    
    FILE* datafile = stdin;
    if (datafilename!="-") datafile = fopen(datafilename.c_str(), "r");
    if (!datafile) {
        cerr << optionsdesc << endl;
        cerr << "Error: Cannot open data file!" << endl;
        return 1;
    }

    vector<vector<double>> data;
    int linenum = 0;
    char* line = 0; size_t linelen = 0; int num_read = 0;
    int num_data_per_line = -1;
    while ((num_read = getline(&line, &linelen, datafile)) != -1) {
        ++linenum;
        // ignore comments and eof
        if (linelen==0 || line[0]=='#') continue;
        // ignore blank lines
        char* x = line;
        while ((*x==' ')||(*x=='\t')||(*x=='\n')||(*x=='\r')) ++x; 
        if (*x==0) continue;
        data.resize(data.size()+1);
        while(*x!=0) {
            double value = fast_atof_next_token(x);
            data.back().push_back(value);
        }
        if (num_data_per_line==-1) num_data_per_line = data.back().size();
        else if (num_data_per_line != data.back().size()) {
            if (verbose) cout << "Warning: Inconsistent number of data values line " << linenum << " (" << data.back().size() << ", should be " << num_data_per_line << ")" << endl;
            // ensure no out-of-range condition even in this case
            if (data.back().size()<num_data_per_line) data.back().resize(num_data_per_line);
        }
    }
    fclose(datafile);
    
    if (num_data_per_line<3) {
        cerr << optionsdesc << endl;
        cerr << "Error: Invalid data file: need at least 3 columns per line." << endl;
        return 1;
    }

    if (fields.empty()) {
        if (num_data_per_line==3) fields="1:2:3";
        else if (num_data_per_line==4) fields="1:2:3:4";
        else fields="1:2:3:4:5";
    }
    typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
    boost::char_separator<char> separator(":");
    tokenizer columntokens(fields, separator);
    vector<int> column;
    for (tokenizer::iterator it = columntokens.begin(); it != columntokens.end(); ++it) {
        column.push_back(atoi(it->c_str())-1); // account for 1->0 for the indexing
        if (column.back()<0 || column.back()>=num_data_per_line) {
            cerr << optionsdesc << endl;
            cerr << "Error: Invalid column number in the --using argument." << endl;
            return 1;
        }
    }
    
    if (column.size()<3 || column.size()>5) {
        cerr << optionsdesc << endl;
        cerr << "Error: Invalid columns specification in the --using argument, need between 3 and 5 values." << endl;
        return 1;
    }
    int x = column[0];
    int y = column[1];
    int vx = -1, vy = -1, s = -1;
    if (column.size()==3) s = column[2];
    else {
        vx=column[2]; 
        vy=column[3];
        if (column.size()==5) s = column[4];
    }
    
    if (!xrange.empty()) {
        tokenizer minmaxtokens(xrange, separator);
        tokenizer::iterator it = minmaxtokens.begin();
        double dmin, dmax;
        bool hasrange = true;
        if (it != minmaxtokens.end()) dmin = atof((*it++).c_str());
        else hasrange = false;
        if (it != minmaxtokens.end()) dmax = atof((*it++).c_str());
        else hasrange = false;
        if (hasrange) {
            int ndata = data.size();
            for (int i=0; i<ndata;) {
                if (data[i][x]<dmin || data[i][x]>dmax) {
                    data[i].swap(data[--ndata]);
                }
                else ++i;
            }
            data.resize(ndata);
        } else {
            if (verbose) cout << "Warning: invalid xrange specified, ignored." << endl;
        }
    }
    
    if (!yrange.empty()) {
        tokenizer minmaxtokens(yrange, separator);
        tokenizer::iterator it = minmaxtokens.begin();
        double dmin, dmax;
        bool hasrange = true;
        if (it != minmaxtokens.end()) dmin = atof((*it++).c_str());
        else hasrange = false;
        if (it != minmaxtokens.end()) dmax = atof((*it++).c_str());
        else hasrange = false;
        if (hasrange) {
            int ndata = data.size();
            for (int i=0; i<ndata;) {
                if (data[i][y]<dmin || data[i][y]>dmax) {
                    data[i].swap(data[--ndata]);
                }
                else ++i;
            }
            data.resize(ndata);
        } else {
            if (verbose) cout << "Warning: invalid yrange specified, ignored." << endl;
        }
    }

    set<double> xvalues, yvalues;
    double minx = numeric_limits<double>::max();
    double maxx = -numeric_limits<double>::max();
    double miny = numeric_limits<double>::max();
    double maxy = -numeric_limits<double>::max();
    for(auto v : data) {
        xvalues.insert(v[x]);
        minx = min(minx,v[x]);
        maxx = max(maxx,v[x]);
        yvalues.insert(v[y]);
        miny = min(miny,v[y]);
        maxy = max(maxy,v[y]);
    }

    int nx = xvalues.size();
    int ny = yvalues.size();

    // outer boundaries: from minx to maxx at the pixels boundaries
    // maxx included in the last pixel
    // default mode: one pixel per distinct value, assume regularly spaced
    // minx is at the center of the first pixel, maxx also
    // | minx | ... | ... | maxx |
    // ex: 4 pixels, nx=4, with minx=18, maxx=24, maxx-minx=6.
    // (maxx-minx) = dx * (nx-1);
    double dx = 1;
    double dy = 1;
    if (nx>1) dx = (maxx-minx) / (nx-1);
    if (ny>1) dy = (maxy-miny) / (ny-1);
    
    ImageProducer ip(nx*zoomfactor,ny*zoomfactor,dx/zoomfactor,dy/zoomfactor,minx-(outer_boundaries?0:(0.5*dx)),miny-(outer_boundaries?0:(0.5*dy)));
    for(auto v : data) {
        int xi = max(0,min(nx-1,(int)round((v[x]-minx)/dx)));
        int yi = max(0,min(ny-1,(int)round((v[y]-miny)/dy)));
        for (int zxi = 0; zxi < zoomfactor; ++zxi) for (int zyi = 0; zyi < zoomfactor; ++zyi)
        if (vx==-1) ip.plot(xi*zoomfactor+zxi,yi*zoomfactor+zyi,v[s]);
        else if (s==-1) ip.vfplot(xi*zoomfactor+zxi,yi*zoomfactor+zyi,v[vx],v[vy]);
        else ip.vfplot(xi*zoomfactor+zxi,yi*zoomfactor+zyi,v[vx],v[vy],v[s]);
    }


    if (!arrows_range.empty()) {
        tokenizer minmaxtokens(arrows_range, separator);
        tokenizer::iterator it = minmaxtokens.begin();
        double dmin, dmax;
        bool hasrange = true;
        if (it != minmaxtokens.end()) dmin = atof((*it++).c_str());
        else hasrange = false;
        if (it != minmaxtokens.end()) dmax = atof((*it++).c_str());
        else hasrange = false;
        if (hasrange) ip.set_arrows_range(dmin,dmax);
        else {
            if (verbose) cout << "Warning: invalid arrows range specified, using median value from the data" << endl;
        }
    }
    
    if (scalar_range.empty()) ip.save(outputfilename);
    else {
        tokenizer minmaxtokens(scalar_range, separator);
        tokenizer::iterator it = minmaxtokens.begin();
        double dmin, dmax;
        bool hasrange = true;
        if (it != minmaxtokens.end()) dmin = atof((*it++).c_str());
        else hasrange = false;
        if (it != minmaxtokens.end()) dmax = atof((*it++).c_str());
        else hasrange = false;
        if (hasrange) ip.save(outputfilename,dmin,dmax);
        else {
            if (verbose) cout << "Warning: invalid scalar range specified, using min/max from the data" << endl;
            ip.save(outputfilename);
        }
    }

    return 0;
}
